import com.gpudb.*;
import com.gpudb.protocol.*;

import java.util.*;
import java.util.concurrent.*;

import java.time.*;


/*
 * Multi-Head Ingress/Egress with the Java API
 *
 * This example will demonstrate both multi-head ingestion and multi-head egress
 * (keyed lookup).
 * 
 * It makes use of an order_history table for ingestion and a store_sales
 * materialized view for egress, which need to be created beforehand.
 */
public class MultiHeadUseCase
{
	private static final String SCHEMA_NAME = "tutorial_mh_java";
	private static final int TOTAL_STORES = 10;


	/*
	 * Helper class for parallelizing usage of the bulk ingest object.
	 * 
	 * This class will generate a batch of inserts of a given size, starting at
	 * a given index, and then add them to a given BulkInserter
	 */
	private static class BatchInsert implements Runnable
	{
		/* Bulk ingestion object to use for inserting records*/
		BulkInserter<GenericRecord> bulkInserter;
		/* Database type schema of records being inserted */
		Type type;
		/* ID/index of first record */
		int startIndex;
		/* Number of records to insert */
		int batchSize;

		/*
		 * Creates a new batch insert object, which will generate objects of the
		 * specified type and add them to the specified bulk ingest object
		 * 
		 * @param bulkInserter bulk ingest object to which records will be added
		 * @param type database type schema to use for objects being inserted
		 * @param startIndex starting index of records being inserted; will be
		 *        used as order ID for inserted records
		 * @param batchSize number of records to insert in this batch
		 */
		public BatchInsert(BulkInserter<GenericRecord> bulkInserter, Type type, int startIndex, int batchSize)
		{
			this.bulkInserter = bulkInserter;
			this.type = type;
			this.startIndex = startIndex;
			this.batchSize = batchSize;
		}

		public void run()
		{
			try
			{
				int endIndex = startIndex + batchSize;

				for (int recIndex = startIndex; recIndex < endIndex; recIndex++)
				{
					// Generate data
					final int orderId = recIndex;
					final int storeId = ThreadLocalRandom.current().nextInt(1, TOTAL_STORES + 1);
					final double totalAmount = ThreadLocalRandom.current().nextDouble(0, 1000) + .01;
					final long timestamp = LocalDateTime.now().toEpochSecond(ZoneOffset.UTC) * 1000;

					// Populate record
					GenericRecord order = new GenericRecord(type);
					order.put(0, storeId);
					order.put(1, orderId);
					order.put(2, totalAmount);
					order.put(3, timestamp);
					bulkInserter.insert(order);
				}
			}
			catch (Exception e)
			{
				System.out.println("Error inserting record: " + e);
			}
		}
	}

	/*
	 * Creates the schema and tutorial objects
	 * 
	 * @param db Kinetica database connection to use for creating objects
	 */
	private static void setup(GPUdb db) throws GPUdbException
	{
		// (Re)create the tutorial schema
		Map<String, String> options = GPUdb.options(
				DropSchemaRequest.Options.NO_ERROR_IF_NOT_EXISTS, DropSchemaRequest.Options.TRUE,
				DropSchemaRequest.Options.CASCADE, DropSchemaRequest.Options.TRUE
		);
		db.dropSchema(SCHEMA_NAME, options);
		db.createSchema(SCHEMA_NAME, null);

		// Create the order history table
		String ddlOrderHistory = "" +
		"   CREATE TABLE " + SCHEMA_NAME + ".order_history    " +
		"    (                                                " +
		"        store_id INT NOT NULL,                       " +
		"        id INT NOT NULL,                             " +
		"        total_amount DOUBLE NOT NULL,                " +
		"        timestamp TYPE_TIMESTAMP NOT NULL,           " +
		"        PRIMARY KEY (id, store_id),                  " +
		"        SHARD KEY (store_id)                         " +
		"    );                                               ";
		db.executeSql(new ExecuteSqlRequest().setStatement(ddlOrderHistory));

		// Create the store sales materialized view
		String ddlStoreSales = "" +
		"    CREATE MATERIALIZED VIEW " + SCHEMA_NAME + ".store_sales    " +
		"    REFRESH EVERY 10 MINUTES AS                                 " +
		"    SELECT  /* KI_HINT_GROUP_BY_PK */                           " +
		"        store_id,                                               " +
		"        DATE(timestamp) AS order_date,                          " +
		"        SUM(total_amount) AS total_sales                        " +
		"    FROM                                                        " +
		"        " + SCHEMA_NAME + ".order_history                       " +
		"    GROUP BY                                                    " +
		"        store_id,                                               " +
		"        DATE(timestamp);                                        ";
		db.executeSql(new ExecuteSqlRequest().setStatement(ddlStoreSales));
	}

	/*
	 * Performs a bulk ingestion of data into Kinetica
	 * 
	 * Creates a BulkInserter and 4 record-generating threads to push a
	 * pre-configured amount of data into the database.  The target shard for
	 * each record will be calculated on the client side, and records will be
	 * pushed directly to the database ranks serving the respective shards in
	 * batches of preconfigured size.
	 * 
	 * Steps required for bulk ingestion:
	 * 
	 * 1. Acquire a handle to the database
	 * 2. Retrieve the type schema of the target table from the database
	 * 3. Create a BulkInserter for the table with the retrieved type schema
	 * 4. Insert records with the BulkInserter, flushing its queues when done
	 * 
	 * NOTE:  The BulkInserter has no inherent parallelism of its own, but is
	 *        thread-safe, and can be made more performant by having multiple
	 *        threads use it.  In this case, for example, using 4 threads allows
	 *        4 insert requests to execute concurrently.
	 * 
	 * @param db Kinetica database connection to use for ingestion
	 */
	private static void ingest(GPUdb db) throws GPUdbException
	{
		final String tableName = SCHEMA_NAME + ".order_history";
		final int totalThreads = 4;
		final int totalBatches = 10;
		final int batchSize = 100000;
		final int queueSize = 10000;


		// Create database objects before ingesting
		setup(db);

		// Acquire a type for the order history table
		Type orderType = Type.fromTable(db, tableName);

		// Acquire a bulk inserter to perform the multi-head ingest
		BulkInserter<GenericRecord> bulkInserter =
				new BulkInserter<>(db, tableName, orderType, queueSize, null, new WorkerList(db));

		// Create a thread pool for generating & inserting data
		ExecutorService executorService = Executors.newFixedThreadPool(totalThreads);

		for (int batchIndex = 0; batchIndex < totalBatches; batchIndex++)
			executorService.execute
			(
					new BatchInsert(bulkInserter, orderType, batchIndex * batchSize, batchSize)
			);

		// Shut down thread pool
		executorService.shutdown();
		try
		{
			if (!executorService.awaitTermination(5, TimeUnit.SECONDS))
				executorService.shutdownNow();
		}
		catch (InterruptedException e)
		{
			executorService.shutdownNow();
		}

		// To ensure all records are inserted, flush the bulk inserter object
		bulkInserter.flush();
	}

	/*
	 * Performs a keyed extraction of data from Kinetica
	 * 
	 * Creates a RecordRetriever to look up sales totals of each store in the
	 * database for the current day.  The sales totals are aggregated in the
	 * materialized view used as the egress source.  The key for lookup is the
	 * store ID and the filter applied is for the current date.
	 * 
	 * Steps required for keyed lookup:
	 * 
	 * 1. Acquire a handle to the database
	 * 2. Retrieve the type schema of the source table from the database
	 * 3. Create a RecordRetriever for the table with the retrieved type schema
	 * 4. Lookup records by key with the RecordRetriever
	 * 
	 * @param db Kinetica database connection to use for keyed lookup
	 */
	private static void egress(GPUdb db) throws GPUdbException
	{
		final String viewName = SCHEMA_NAME + ".store_sales";
		final String today = LocalDate.now().toString();
		final String todayFilter = "order_date = '" + today + "'";

		// Acquire a type for the store total table
		Type storeTotalType = Type.fromTable(db, viewName);

		// Acquire a record retriever to perform the multi-head egress
		RecordRetriever<GenericRecord> recordRetriever =
				new RecordRetriever<>(db, viewName, storeTotalType, new WorkerList(db));

		System.out.println("Summary for date: [" + today + "]");
		System.out.println("Store #  Total Sales");
		System.out.println("======= =============");

		for (int storeNum = 1; storeNum <= TOTAL_STORES; storeNum++)
		{
			// Set up store number to look up
			List<Object> storeKeyset = GPUdb.list((Object)storeNum);

			// Request filtered records
			GetRecordsResponse<GenericRecord> grResp = recordRetriever.getByKey(storeKeyset, todayFilter);

			// Output columns 1 and 3 from table--store ID and total sales
			for (GenericRecord storeTotal : grResp.getData())
				System.out.printf("%7d %13.2f%n", storeTotal.get(0), storeTotal.get(2));
		}
	}

	/*
	 * Refreshes the materialized view that is the source for multi-head egress
	 * 
	 * @param db Kinetica database connection to use for refreshing materialized
	 *        view
	 */
	private static void refresh(GPUdb db) throws GPUdbException
	{
		// Refresh the egress source materialized view
		db.alterTable(SCHEMA_NAME + ".store_sales", AlterTableRequest.Action.REFRESH, null, null);
	}


	/*
	 * Runs the multi-head ingest/egress example
	 * 
	 * Commands:
	 * 
	 * RUN     - run an ingest, materialized view refresh, and egress
	 * INGEST  - run only the ingest portion of the example
	 * EGRESS  - run only the egress portion of the example
	 * REFRESH - refresh the materialized view
	 * 
	 * An optional host can be specified for non-local database instances.  A
	 * username & password can also be specified for authentication.
	 * 
	 * NOTE:  In the common implementation of the use case, the materialized
	 * view would be set to periodically refresh.  This example will force a
	 * refresh after ingest to alleviate the need to wait for a refresh cycle.
	 */
	public static void main(String[] args) throws GPUdbException
	{
		if (args.length == 0)
		{
			System.out.println("No command specified; options: <RUN|INGEST|REFRESH|EGRESS> <host> [<user> <pass>]");
			System.exit(-1);
		}

		String command = args[0];
		String host = (args.length > 1) ? args[1] : "localhost";
		String user = (args.length > 2) ? args[2] : null;
		String pass = (args.length > 3) ? args[3] : null;

		GPUdbBase.Options options = new GPUdbBase.Options();
		options.setUsername(user);
		options.setPassword(pass);
		GPUdb db = new GPUdb("http://" + host + ":9191", options);

		if ("RUN".equalsIgnoreCase(command))
		{
			ingest(db);
			refresh(db);
			egress(db);
		}
		else if ("INGEST".equalsIgnoreCase(command))
			ingest(db);
		else if ("REFRESH".equalsIgnoreCase(command))
			refresh(db);
		else if ("EGRESS".equalsIgnoreCase(command))
			egress(db);
		else
		{
			System.out.println("Command [" + command + "] not found; options: <RUN|INGEST|REFRESH|EGRESS> <host> [<user> <pass>]");
			System.exit(-2);
		}
		
		System.exit(0);
	}
}
